(******************************************************************************
 * Copyright (c) 2012, Toyama&Aoto Laboratory, Tohoku University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the 
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(******************************************************************************
 * file: rwtools/util/list_util.sml
 * description: utility for list functions
 * author: AOTO Takahito
 * 
 * $Id: list_util.sml,v 1.7 2012/05/14 04:51:59 aoto Exp $
 ******************************************************************************)

signature LIST_UTIL =  
sig
    val mapAppend: ('a -> 'b list) -> 'a list -> 'b list
    val replaceNth: ('a list * int * 'a) -> 'a list
    val exceptNth: ('a list * int) -> 'a list

    val member: ''a -> ''a list -> bool
    val member': ('a * 'a -> bool) -> 'a -> 'a list -> bool

    val union: (''a list * ''a list) -> ''a list
    val union': ('a * 'a -> bool) -> ('a list * 'a list) -> 'a list

    val indexOf: ''a -> ''a list -> int option
    val indexOf': ('a * 'a -> bool) -> 'a -> 'a list -> int option

    val deleteOne: ''a -> ''a list -> ''a list 
    val deleteAll: ''a -> ''a list -> ''a list 
    val differenceByOne: ''a list * ''a list -> ''a list 
    val differenceByAll: ''a list * ''a list -> ''a list 

    val deleteOne': ('a * 'a -> bool) -> 'a -> 'a list -> 'a list
    val deleteAll': ('a * 'a -> bool) -> 'a -> 'a list -> 'a list
    val differenceByOne': ('a * 'a -> bool) -> 'a list * 'a list -> 'a list
    val differenceByAll': ('a * 'a -> bool) -> 'a list * 'a list -> 'a list

    val eliminateDuplication: ''a list -> ''a list
    val eliminateDuplication': ('a * 'a -> bool) -> 'a list -> 'a list

    val toString: ('a -> string) -> string * string * string -> 'a list -> string
    val toStringComma: ('a -> string) -> 'a list -> string
    val toStringSpace: ('a -> string) -> 'a list -> string
    val toStringAst: ('a -> string) -> 'a list -> string
    val toStringPlus: ('a -> string) -> 'a list -> string
    val toStringPlusRound: ('a -> string) -> 'a list -> string
    val toStringSlashSquare: ('a -> string) -> 'a list -> string

    val toStringCommaSquare: ('a -> string) -> 'a list -> string
    val toStringCommaAngle: ('a -> string) -> 'a list -> string
    val toStringCommaCurly: ('a -> string) -> 'a list -> string
    val toStringCommaRound: ('a -> string) -> 'a list -> string
    val toStringCommaLnSquare: ('a -> string) -> 'a list -> string
    val toStringCommaLnCurly: ('a -> string) -> 'a list -> string

    val toStringSpaceSquare: ('a -> string) -> 'a list -> string
    val toStringSpaceAngle: ('a -> string) -> 'a list -> string
    val toStringSpaceCurly: ('a -> string) -> 'a list -> string
    val toStringSpaceRound: ('a -> string) -> 'a list -> string
    val toStringSpaceLnSquare: ('a -> string) -> 'a list -> string
    val toStringSpaceLnCurly: ('a -> string) -> 'a list -> string

    val toStringSemicolonSquare: ('a -> string) -> 'a list -> string
    val toStringSemicolonAngle: ('a -> string) -> 'a list -> string
    val toStringSemicolonCurly: ('a -> string) -> 'a list -> string
    val toStringSemicolonRound: ('a -> string) -> 'a list -> string
    val toStringSemicolonLnSquare: ('a -> string) -> 'a list -> string
    val toStringSemicolonLnCurly: ('a -> string) -> 'a list -> string

    val powerlist: 'a list -> 'a list list
    val powerlistWithComplement: 'a list -> ('a list * 'a list) list
end

structure ListUtil: LIST_UTIL =
struct
  local 
  structure L = List

  in 

  fun mapAppend f xs = L.foldr (fn (x,ys) => L.@(f x, ys)) [] xs

    fun replaceNth (xs,i,y) = (L.take (xs,i)) @ [y] @ (L.drop (xs,i+1))

    fun exceptNth ([],n) = []
      | exceptNth (x::xs,n) = if n = 0 then xs else (x::(exceptNth (xs,n-1)))

   (* ys $B$NCf$K(B x $B$,$"$k$+!)(B *)
    fun member x ys = L.exists (fn y => x = y) ys

   (* ys $B$NCf$K(B x $B$,$"$k$+!)(B *)
    fun member' pred x ys = L.exists (fn y => pred (x,y)) ys

   (* xs $B$H(B ys $B$r!$=EJ#$r=|$$$F9gJ;(B *)
    fun union (xs,ys) = 
	let fun unionSub [] ys ans = L.revAppend (ans,ys)
	      | unionSub (x::xs) ys ans = 
		if member x ys
		then unionSub xs ys ans
		else unionSub xs ys (x::ans)
	in unionSub xs ys []
	end

   (* xs $B$H(B ys $B$r!$=EJ#$r=|$$$F9gJ;(B *)
    fun union' pred (xs,ys) = 
	let fun unionSub [] ys ans = L.revAppend (ans,ys)
	      | unionSub (x::xs) ys ans = 
		if member' pred x ys
		then unionSub xs ys ans
		else unionSub xs ys (x::ans)
	in unionSub xs ys []
	end

   (* x $B$,(B ys $B$NCf$G2?HVL\$NMWAG$H$7$F=P8=$9$k$+$r(B option $B7?$GJV$9(B *)
   (* $B@hF,$J$i(B 0$BHVL\!"=P8=$7$J$$$H$-$O(B NONE *)
    fun indexOf x ys = 
	let fun indexOfSub [] _= NONE
	      | indexOfSub (z::zs) num = 
		if x = z then SOME num else indexOfSub zs (1+num)
	in indexOfSub ys 0 
	end

    fun indexOf' pred x ys = 
	let fun indexOfSub [] _= NONE
	      | indexOfSub (z::zs) num = 
		if pred (x,z) then SOME num else indexOfSub zs (1+num)
	in indexOfSub ys 0 
	end

   (* ys $B$NCf$+$i(B x $B$r(B1$B$D:o=|(B($B$"$l$P(B)$B$9$k(B *)
    fun deleteOne x ys = 
	let fun delete [] ans = rev ans
	      | delete (y'::ys') ans = 
		if x = y' then L.revAppend (ans, ys')
		else delete ys' (y'::ans)
	in delete ys [] 
	end

   (* ys $B$NCf$+$i(B x $B$r$9$Y$F:o=|$9$k(B *)
    fun deleteAll x ys = 
	let fun delete [] ans = rev ans
	      | delete (y'::ys') ans = 
		if x = y' then delete ys' ans
		else delete ys' (y'::ans)
	in delete ys [] 
	end

   (* xs $B$NCf$+$i(B ys $B$HF1$8MWAG$rF1$88D?t$@$1:o=|$9$k(B *)
    fun differenceByOne (xs,[]) = xs
      | differenceByOne (xs,y::ys) = differenceByOne (deleteOne y xs, ys)

   (* xs $B$NCf$+$i!$(Bys $B$K$b$"$kMWAG$r!$A4It:o=|$9$k(B *)
    fun differenceByAll (xs,[]) = xs
      | differenceByAll (xs,y::ys) = differenceByAll (deleteAll y xs, ys)

   (* ys $B$NCf$+$i(B x $B$r(B1$B$D:o=|(B($B$"$l$P(B)$B$9$k(B *)
    fun deleteOne' pred x ys = 
	let fun delete [] ans = rev ans
	      | delete (y'::ys') ans = 
		if pred (x,y') then L.revAppend (ans, ys')
		else delete ys' (y'::ans)
	in delete ys [] 
	end

   (* ys $B$NCf$+$i(B x $B$r$9$Y$F:o=|$9$k(B *)
    fun deleteAll' pred x ys = 
	let fun delete [] ans = rev ans
	      | delete (y'::ys') ans = 
		if pred (x,y') then delete ys' ans
		else delete ys' (y'::ans)
	in delete ys [] 
	end

   (* xs $B$NCf$+$i(B ys $B$HF1$8MWAG$rF1$88D?t$@$1:o=|$9$k(B *)
    fun differenceByOne' _ (xs,[]) = xs
      | differenceByOne' pred (xs,y::ys) = differenceByOne' pred (deleteOne' pred y xs, ys)

   (* xs $B$NCf$+$i!$(Bys $B$K$b$"$kMWAG$r!$A4It:o=|$9$k(B *)
    fun differenceByAll' _  (xs,[]) = xs
      | differenceByAll' pred (xs,y::ys) = differenceByAll' pred (deleteAll' pred y xs, ys)

    fun eliminateDuplication xs = 
	let fun elimDup [] ans = ans
	      | elimDup (x::xs) ans = 
		if L.exists (fn y => x = y) ans
		then elimDup xs ans
		else elimDup xs (x::ans)
	in elimDup xs []
	end

    fun eliminateDuplication' pred xs = 
	let fun elimDup [] ans = ans
	      | elimDup (x::xs) ans = 
		if L.exists (fn y => pred (x,y)) ans
		then elimDup xs ans
		else elimDup xs (x::ans)
	in elimDup xs []
	end


    fun toString prElm (start,sep,stop) xs =
	let fun toStringSub [] = ""
	      | toStringSub [x] = (prElm x)
	      | toStringSub (x::xs)= (prElm x) ^ sep ^ (toStringSub xs)
	in  start ^ (toStringSub xs) ^ stop
	end

    fun toStringComma prElm xs = toString prElm ("",  ",", "") xs
    fun toStringSpace prElm xs = toString prElm ("",  " ", "") xs
    fun toStringAst prElm xs = toString prElm   ("",  "*", "") xs
    fun toStringPlus prElm xs = toString prElm   ("",  "+", "") xs
    fun toStringPlusRound prElm xs = toString prElm   ("(",  "+", ")") xs
    fun toStringSlashSquare prElm xs = toString prElm   ("[",  "/", "]") xs

    fun toStringSpaceSquare prElm xs = toString prElm ("[",  " ",  "]") xs
    fun toStringSpaceAngle prElm xs = toString prElm  ("<",  " ",  ">") xs
    fun toStringSpaceCurly prElm xs = toString prElm  ("{",  " ",  "}") xs
    fun toStringSpaceRound prElm xs = toString prElm  ("(",  " ",  ")") xs
    fun toStringSpaceLnSquare prElm xs = toString prElm ("   [ ",  "\n     ",  " ]\n") xs
    fun toStringSpaceLnCurly prElm xs = toString prElm ("   { ",  "\n     ",  " }\n") xs


    fun toStringCommaSquare prElm xs = toString prElm ("[",  ",",  "]") xs
    fun toStringCommaAngle prElm xs = toString prElm  ("<",  ",",  ">") xs
    fun toStringCommaCurly prElm xs = toString prElm  ("{",  ",",  "}") xs
    fun toStringCommaRound prElm xs = toString prElm  ("(",  ",",  ")") xs
    fun toStringCommaLnSquare prElm xs = toString prElm ("   [ ",  ",\n     ",  " ]\n") xs
    fun toStringCommaLnCurly prElm xs = toString prElm ("   { ",  ",\n     ",  " }\n") xs

    fun toStringSemicolonSquare prElm xs = toString prElm ("[",  ";",  "]") xs
    fun toStringSemicolonAngle prElm xs = toString prElm  ("<",  ";",  ">") xs
    fun toStringSemicolonCurly prElm xs = toString prElm  ("{",  ";",  "}") xs
    fun toStringSemicolonRound prElm xs = toString prElm  ("(",  ";",  ")") xs
    fun toStringSemicolonLnSquare prElm xs = toString prElm ("   [ ",  ";\n     ",  " ]\n") xs
    fun toStringSemicolonLnCurly prElm xs = toString prElm ("   { ",  ";\n     ",  " }\n") xs


    (* powerlist: [0,1,2] 
     =>  [[],[0],[1],[2],[0,1],[0,2],[1,2],[0,1,2]]  *)
    fun powerlist [] = [[]]
      | powerlist (x::xs) = 
	let val yss = powerlist xs
	in yss @ (L.map (fn ys => x::ys) yss)
	end

    (* powerlistWithComplement: 
     [0,1,2]  => 
     [([1,2,3],[]),
      ([1,2],[3]),([1,3],[2]),([1],[2,3]),
      ([2,3],[1]),([2],[1,3]),([3],[1,2]),([],[1,2,3])] *)
    fun powerlistWithComplement [] = [([],[])]
      | powerlistWithComplement (x::xs) = 
	let val yss = powerlistWithComplement xs
	in (L.map (fn (ys,zs) => (x::ys,zs)) yss)
	   @ (L.map (fn (ys,zs) => (ys,x::zs)) yss)
	end


  end
end
